package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.model.SourceType.FAULT;
import static gov.usgs.earthquake.nshmp.model.SourceType.INTERFACE;
import static java.util.stream.Collectors.toList;

import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.function.Predicate;

import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Type;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Cluster source representation. Each cluster source wraps a {@code RuptureSet}
 * containing one or more fault representations that rupture as independent
 * events but with a similar rate. For example, at New Madrid, each
 * ClusterRuptureSet has 5 ClusterSources, one for each position variant of the
 * model. For each position variant there is one FaultRuptureSet containing the
 * FaultSources in the cluster, each of which may have one, or more, magnitude
 * or other variants represented by its internal list of {@code Mfd}s.
 *
 * <p>Cluster source hazard is calculated from the joint probabilities of ground
 * motions from the wrapped faults. It is only used for finite fault and
 * interface sources.
 *
 * <p>A {@code ClusterSource} cannot be created directly; it may only be created
 * by a private parser.
 *
 * @author U.S. Geological Survey
 */
public class ClusterRuptureSet extends AbstractRuptureSet<ClusterSource> {

  /*
   * Developer notes:
   *
   * A ClusterRuptureSet is composed of two or more FaultRuptureSets (fault
   * sections) that rupture together according to some rate or logic tree of
   * rates. Each nested FaultRuptureSet itself contains a logic tree of SINGLE
   * MFDs, usually labelled 'M1', 'M2', etc. The MFDs for each FaultRuptureSet
   * are the same size and have identical weights. Technically, the M1 fault
   * branches across each cluster should be considered together, the M2 fault
   * branches together, etc. in the cluster hazard calculation. However, the
   * magnitude branch exceedance curves are currently collapsed prior to doing
   * the cluster calculation.
   *
   * For the purposes of reporting an MFD tree for a ClusterRuptureSet, we
   * create MFDs for each magnitude branch spanning the individual sections in
   * the cluster. In some cases there may be identical magnitudes in a cluster
   * such that the cluster rate reflected in the MFD may be doubled or more.
   *
   */

  @Deprecated
  final ModelData data; // holds cluster rate tree
  private final List<ClusterSource> source;
  private final LogicTree<Mfd> mfdTree;

  private ClusterRuptureSet(Builder builder) {
    super(builder);
    this.data = builder.data;
    this.source = List.of(builder.source);
    this.mfdTree = builder.mfdTree;
  }

  @Override
  public List<Feature> features() {
    return source.get(0).ruptureSets.stream()
        .map(RuptureSet::features)
        .flatMap(List::stream)
        .collect(toList());
  }

  @Override
  public int size() {
    return 1; // faultRuptureSets.size();
  }

  @Override
  public ClusterSource get(int index) {
    return source.get(index);
  }

  @Override
  public LogicTree<Mfd> mfdTree() {
    return mfdTree;
  }

  /*
   * With the switch to RuptureSet implementing List, when we add a rupture set
   * to a Map, hashCode() is called on the contents of the rupture set. Cluster
   * sources have special handling and so we revert to the system hash code and
   * equals. See also SystemRuptureSet.
   */

  @Override
  public int hashCode() {
    return System.identityHashCode(this);
  }

  @Override
  public boolean equals(Object obj) {
    return (this == obj);
  }

  @Override
  public Iterator<ClusterSource> iterator() {
    return source.iterator();
  }

  @Override
  public Predicate<ClusterSource> distanceFilter(
      final Location loc,
      final double distance) {

    return new Predicate<ClusterSource>() {

      private final Predicate<FaultSource> faultFilter =
          new FaultRuptureSet.DistanceFilter(loc, distance);
      private final Predicate<InterfaceSource> interfaceFilter =
          new InterfaceRuptureSet.DistanceFilter(loc, distance);

      @Override
      public boolean test(ClusterSource cs) {
        SourceType type = cs.ruptureSets().get(0).type();
        if (type == FAULT) {
          return cs.ruptureSets().stream()
              .map(rs -> rs.iterator().next())
              .map(source -> (FaultSource) source)
              .anyMatch(faultFilter);

        } else if (type == INTERFACE) {
          return cs.ruptureSets().stream()
              .map(rs -> rs.iterator().next())
              .map(source -> (InterfaceSource) source)
              .anyMatch(interfaceFilter);

        } else {
          throw new IllegalStateException("Invalid cluster sources type: " + type);
        }
        // Predicate<? extends Source> filter = (type == FAULT)
        // ? faultFilter
        // : interfac eFilter;
        //
        // // this knows that there shouldn't be floating ruptures and so
        // // taking the first rupture will do an adequate geometry test
        // return cs.ruptureSets().stream()
        // .map(rs -> rs.iterator().next())
        // .map(source -> (FaultSource) source)
        // .anyMatch(faultFilter);
      }
    };
  }

  /**
   * The closest point across the traces of all fault sources that participate
   * in this cluster, relative to the supplied site {@code Location}.
   */
  @Override
  public Location location(Location site) {
    return source.get(0).location(site);
  }

  static Builder builder() {
    return new Builder();
  }

  /* Single use builder */
  static class Builder extends AbstractRuptureSet.Builder {

    private ModelData data; // holds cluster rate-tree
    private ClusterSource source;
    private LogicTree<Mfd> mfdTree;

    private Builder() {
      super(SourceType.FAULT_CLUSTER);
    }

    @Override
    Builder name(String name) {
      super.name(name);
      return this;
    }

    @Override
    Builder id(int id) {
      super.id(id);
      return this;
    }

    /* Set MFD helper data. */
    Builder modelData(ModelData data) {
      this.data = checkNotNull(data);
      super.setting(data.tectonicSetting());
      super.gmms(data.gmms());
      return this;
    }

    Builder setSource(ClusterSource source) {
      checkState(this.source == null);
      this.source = checkNotNull(source);
      return this;
    }

    void validateAndInit(String label) {
      validateState(label);
      checkNotNull(data, "%s model data", label);
      checkNotNull(source, "%s cluster source", label);
    }

    LogicTree<Mfd> createMfdTree() {
      LogicTree.Builder<Mfd> builder = LogicTree.builder("cluster-mfd");
      LogicTree<Mfd> model = source.ruptureSets().get(0).mfdTree();
      /* loop magnitude branches */
      for (int i = 0; i < model.size(); i++) {
        /* collect cluster magnitudes and rate; all M have cluster rate */
        Map<Double, Double> magMap = new TreeMap<>();
        String magId = model.get(i).id();
        double magWeight = model.get(i).weight();
        for (RuptureSet<? extends Source> ruptureSet : source.ruptureSets()) {
          Branch<Mfd> mfd = ruptureSet.mfdTree().get(i);
          /* check magnitude branch IDs and weights match */
          checkState(mfd.id().equals(magId));
          checkState(mfd.value().properties().type() == Type.SINGLE);
          magMap.merge(
              mfd.value().data().x(0),
              source.rate(),
              (v1, v2) -> v1 + v2);
        }
        XySequence mfdXy = XySequence.create(
            magMap.keySet(),
            magMap.values());
        Mfd clusterMfd = Mfd.create(mfdXy);
        builder.addBranch(magId, clusterMfd, magWeight);
      }
      return builder.build();
    }

    ClusterRuptureSet build() {
      validateAndInit("ClusterRuptureSet.Builder");
      mfdTree = createMfdTree();
      return new ClusterRuptureSet(this);
    }
  }

  /*
   * Data container class used during deserialization. ModelLoader uses these
   * data to initialize multiple cluster sources having identical fault sources
   * but scaled to unique rates.
   */
  static class Data {

    final String name;
    final int id;
    final ModelData data;
    final List<? extends RuptureSet<? extends Source>> ruptureSets;

    Data(
        String name,
        int id,
        ModelData data,
        List<? extends RuptureSet<? extends Source>> ruptureSets) {

      if (ruptureSets.size() > 1) {
        for (int i = 1; i < ruptureSets.size(); i++) {
          ModelTrees.checkTreeIdsAndWeights(
              ruptureSets.get(0).mfdTree(),
              ruptureSets.get(i).mfdTree());
        }
      }

      this.name = name;
      this.id = id;
      this.data = data;
      this.ruptureSets = ruptureSets;
    }
  }
}
